#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
#
# This script is licensed under GNU GPL version 2.0 or above
# (c) 2024 Antonio J. Delgado
# (c) 2021 Markus Bergholz https://github.com/markuman/markuman.nextcloud/commits?author=markuman
# Based on https://github.com/markuman/markuman.nextcloud/blob/latest/plugins/modules/password.py

"""Nextcloud Password client"""

from __future__ import absolute_import, division, print_function
__metaclass__ = type
import traceback
import json
import sys
import os
import time
import re
import logging
import csv
from xml.dom import minidom
import binascii
import click
import click_config_file
import passpy
import secretstorage
import cryptography.fernet
from cryptography.fernet import Fernet
from yaml import dump
try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper
try:
    import requests
    HAS_REQUESTS_LIB = True
except ImportError:
    HAS_REQUESTS_LIB = False
    IMPORT_ERROR = traceback.format_exc()

try:
    import pysodium
    HAS_PYSODIUM_LIB = True
except ImportError:
    HAS_PYSODIUM_LIB = False
    IMPORT_ERROR = traceback.format_exc()


DIOS_MIO = """<?xml version="1.0"?>
<d:propfind  xmlns:d="DAV:" xmlns:oc="http://owncloud.org/ns" xmlns:nc="http://nextcloud.org/ns">
<d:prop>
        <d:getlastmodified />
        <d:getetag />
        <d:getcontenttype />
        <d:resourcetype />
        <oc:fileid />
        <oc:permissions />
        <oc:size />
        <oc:checksums />
        <oc:favorite />
        <oc:owner-display-name />
        <oc:share-types />
</d:prop>
</d:propfind>
"""

def parameter_spects(spec_arguments):
    '''Specification of parameters???'''
    argument_spec = {
        "host": {
            "required": False,
            "type": 'str'
        },
        "api_token": {
            "required": False,
            "type": "str"
        },
        "user": {
            "required": False,
            "type": "str"
        },
        "ssl_mode": {
            "required": False,
            "type": "str",
            "default": "https"
        }
    }

    return {**argument_spec, **spec_arguments}


def decrypt(encrypted, key):
    '''Decrypt encrypted data'''
    nonce = encrypted[0:pysodium.crypto_secretbox_NONCEBYTES]
    ciphertext = encrypted[pysodium.crypto_secretbox_NONCEBYTES:]
    return pysodium.crypto_secretbox_open(ciphertext, nonce, key)


def decrypt_item(item_key, item, key='label'):
    '''Decrypt single item'''
    if len(item[key]) == 0:
        return ""
    vals = binascii.unhexlify(item[key])
    return decrypt(vals, binascii.unhexlify(item_key)).decode()


class NextcloudHandler:
    '''Handle Nextcloud Password API'''
    def __init__(self, params):
        self.output_format = params.get('output_format', 'json')
        if 'logger' in params:
            self._log = params['logger']
        else:
            self._init_log(params)
        self.http = 'https'
        self.timeout = params.get('timeout', 3)
        self.ssl = True
        self.encryption_pass = None
        self._get_encryption_pass_from_keyring()
        self.cache = {}
        self.cache_filename = os.path.join(
            os.environ['HOME'],
            '.cache',
            'nc_password_client.cache'
        )
        self.field_replacements = {}
        for field_replacement in params['field_replacements']:
            key, value = field_replacement.split(':')
            self.field_replacements[key] = value
        self.debug(
            {
                "action": "Initializing Nextcloud handler",
                "field_replacements": self.field_replacements
            }
        )
        self._read_cache()
        self.session = requests.Session()
        headers = {
            'User-Agent': 'nc_password_client/0.0.2'
        }
        self.session.headers.update(headers)
        proxies = {
            'http': params.get('https_proxy', ''),
            'https': params.get('https_proxy', ''),
        }
        self.session.proxies.update(proxies)
        self.cache_duration = params.get('cache_duration', 300)
        if params.get('ssl_mode') == 'http':
            self.http = 'http'
        elif params.get('ssl_mode') == 'skip':
            self.ssl = False
        elif os.environ.get('NEXTCLOUD_SSL_MODE') == 'http':
            self.http = 'http'
        elif os.environ.get('NEXTCLOUD_SSL_MODE') == 'skip':
            self.ssl = False

        self.x_api_session = None

        self.host = params.get('host') or os.environ.get('NEXTCLOUD_HOST')
        if self.host is None:
            self.error(
                {
                    "action": "check_params",
                    "message": 'Unable to continue. No Nextcloud Host is given.',
                }
            )
            sys.exit(1)

        self.user = params.get('user') or os.environ.get('NEXTCLOUD_USER')
        if self.user is None:
            self.error(
                {
                    "action": "check_params",
                    "message": 'Unable to continue. No Nextcloud User is given.'
                }
            )
            sys.exit(2)

        self.token = params.get('api_token') or os.environ.get('NEXTCLOUD_TOKEN')
        if self.token is None:
            self.error(
                {
                    "action": "check_params",
                    "message": 'Unable to continue. No Nextcloud Token is given.'
                }
            )
            sys.exit(3)

        self.e2e_password = False
        if params.get('cse_password') or os.environ.get('NEXTCLOUD_CSE_PASSWORD'):
            # manage passwords client side encryption
            self.e2e_password = True
            password = params.get('cse_password') or os.environ.get('NEXTCLOUD_CSE_PASSWORD')
            retval = self.request_passwords_session()
            password_salt = binascii.unhexlify(retval['challenge']['salts'][0])
            generic_hash_key = binascii.unhexlify(retval['challenge']['salts'][1])
            password_hash_salt = binascii.unhexlify(retval['challenge']['salts'][2])

            generic_hash = pysodium.crypto_generichash(
                password.encode() + password_salt,
                k=generic_hash_key,
                outlen=pysodium.crypto_generichash_BYTES_MAX
            )
            password_hash = pysodium.crypto_pwhash(
                pysodium.crypto_box_SEEDBYTES,
                generic_hash,
                password_hash_salt,
                pysodium.crypto_pwhash_OPSLIMIT_INTERACTIVE,
                pysodium.crypto_pwhash_MEMLIMIT_INTERACTIVE,
                pysodium.crypto_pwhash_ALG_ARGON2ID13
            )
            # self.cse['keys']['CSEv1r1']
            self.cse = self.open_passwords_session(password_hash.hex())
            # decrypt CSEv1r1 chain
            vals = binascii.unhexlify(self.cse['keys']['CSEv1r1'])
            salt = vals[0:pysodium.crypto_pwhash_SALTBYTES]
            text = vals[pysodium.crypto_pwhash_SALTBYTES:]
            key = pysodium.crypto_pwhash(
                pysodium.crypto_box_SEEDBYTES,
                password,
                salt,
                pysodium.crypto_pwhash_OPSLIMIT_INTERACTIVE,
                pysodium.crypto_pwhash_MEMLIMIT_INTERACTIVE,
                pysodium.crypto_pwhash_ALG_ARGON2ID13
            )
            self.keychain = json.loads(decrypt(text, key))

    def _get_encryption_pass_from_keyring(self):
        connection = secretstorage.dbus_init()
        collection = secretstorage.get_default_collection(connection)
        if collection.is_locked():
            collection.unlock()
        try:
            for item in collection.get_all_items():
                if item.get_label() == 'nc_password_client':
                    self.encryption_pass = item.get_secret()
        except secretstorage.exceptions.ItemNotFoundException:
            self.error(
                {
                    "action": "_get_encryption_pass_from_keyring",
                    "message": "Item not found browsing keyring items",
                }
            )
        if self.encryption_pass is None:
            self.debug(
                {
                    "action": "_get_encryption_pass_from_keyring",
                    "message": "There wasn't an encryption password in the keyring, will generate one"
                }
            )
            self.encryption_pass = Fernet.generate_key()
            attributes = {
                'application': 'nc_password_client',
            }
            item = collection.create_item('nc_password_client', attributes, self.encryption_pass)
        else:
            self.debug(
                {
                    "action": "_get_encryption_pass_from_keyring",
                    "message": "Encryption password obtained from keyring",
                }
            )

    def _read_cache(self):
        if os.path.exists(self.cache_filename):
            self.debug(
                {
                    "action": "_read_cache",
                    "message": "Reading cached passwords"
                }
            )
            with open(self.cache_filename, 'r', encoding='utf-8') as cache_file:
                content = cache_file.read()
                if len(content) != 0:
                    cipher_suite = Fernet(self.encryption_pass)
                    try:
                        self.cache = json.loads(cipher_suite.decrypt(content))
                        self.debug(
                            {
                                "action": "_read_cache",
                                "last_update": self.cache['last_update'],
                                "total_cached_password": len(self.cache['cached_passwords'])
                            }
                        )
                    except cryptography.fernet.InvalidToken as error:
                        self.debug(
                            {
                                "action": "_read_cache",
                                "message": "Fernet token for passwords local cache is invalid, discarding the local cache.",
                            }
                        )
                        self.cache = {
                            "last_update": -1,
                            "cached_passwords": []
                        }

                else:
                    self.debug(
                        {
                            "action": "_read_cache",
                            "message": "The cache file was empty, so initializing cache"
                        }
                    )
                    self.cache = {
                        "last_update": -1,
                        "cached_passwords": []
                    }
        else:
            self.debug(
                {
                    "action": "_read_cache",
                    "message": "There wasn't a cache file, so initializing cache"
                }
            )
            self.cache = {
                "last_update": -1,
                "cached_passwords": []
            }

    def _write_cache(self):
        self.debug(
                {
                    "action": "_write_cache",
                    "message": "Writing cached passwords to file"
                }
            )
        with open(self.cache_filename, 'bw') as cache_file:
            cipher_suite = Fernet(self.encryption_pass)
            encrypted_cache = cipher_suite.encrypt(bytes(json.dumps(self.cache), 'utf-8'))
            cache_file.write(encrypted_cache)

    def _safer_obj(self, obj, fields=None):
        if fields is None:
            fields = ['password', 'token', 'secret', 'pass', 'passwd', 'hash']
        if isinstance(obj, dict):
            for key in obj.keys():
                if isinstance(obj[key], dict) or isinstance(obj[key], list):
                    obj[key] = self._safer_obj(obj[key], fields=fields)
                else:
                    if key in fields:
                        obj[key] = '***'
        if isinstance(obj, list):
            for item in obj:
                item = self._safer_obj(item, fields=fields)
        return obj

    def _output(self, obj, unsafe=False):
        if unsafe:
            out_obj = obj
        else:
            out_obj = self._safer_obj(obj)
        if self.output_format == 'json':
            return json.dumps(out_obj, indent=2)
        if self.output_format == 'yaml':
            return dump(out_obj, Dumper=Dumper)
        output = ''
        if isinstance(out_obj, list):
            output += '-' * os.get_terminal_size(0)[0]
            for item in obj:
                output += f"{self._output(item)}\n"
        elif isinstance(out_obj, dict):
            output += '-' * os.get_terminal_size(0)[0]
            for key in out_obj.keys():
                output += f"{key} = {self._output(out_obj[key])}\n"
        else:
            output += f"{out_obj}\n"
        return output.strip('\n')

    def debug(self, obj, unsafe=False):
        '''Show debug information'''
        self._log.debug(
            self._output(obj, unsafe=unsafe)
        )

    def warning(self, obj, unsafe=False):
        '''Show warning information'''
        self._log.warning(
            self._output(obj, unsafe=unsafe)
        )

    def info(self, obj, unsafe=False):
        '''Show information'''
        self._log.info(
            self._output(obj, unsafe=unsafe)
        )

    def error(self, obj, unsafe=False):
        '''Show error information'''
        try:
            self._log.error(
                self._output(obj, unsafe=unsafe)
            )
        except TypeError as error:
            self._log.error(
                'Additional error showing the error from %s. %s',
                obj,
                error
            )

    def _init_log(self, params):
        ''' Initialize log object '''
        self._log = logging.getLogger("nc_password_client")
        self._log.setLevel(logging.DEBUG)

        streamhandler = logging.StreamHandler(sys.stdout)
        streamhandler.setLevel(
            logging.getLevelName(params.get("debug_level", 'INFO'))
        )
        self._log.addHandler(streamhandler)
        return True

    def headers(self):
        '''Generate headers'''
        headers = {
            'Accept': 'application/json',
            'OCS-APIRequest': 'true'
        }
        if self.x_api_session:
            headers['X-API-SESSION'] = self.x_api_session
        return headers

    def get(self, path):
        '''Do a GET request'''
        self.debug({ "action": "get", "message": f"Requesting {path}" })
        try:
            r = self.session.get(
                f'{self.http}://{self.host}/{path}',
                auth=(self.user, self.token), verify=self.ssl, headers=self.headers(),
                timeout=self.timeout
            )

            if r.status_code < 299:
                self.debug(
                    {"action": "get", "status_code": r.status_code, "size": len(r.content)}
                )
                return r.json()
            if r.status_code == 404:
                self.error(
                    {
                        "action": "get",
                        "message": 'Path does not exist',
                        "object": path
                    }
                )
                sys.exit(3)
            else:
                self.error(
                    {
                        "action": "get",
                        "message": f'Server return error {r.status_code}',
                    }
                )
                sys.exit(4)
        except requests.exceptions.ReadTimeout as error:
            self.error(
                {
                    "action": "get",
                    "message": f"Timeout ({self.timeout} sec) error doing GET request. %s",
                    "error": f"{error}"
                }
            )
            # sys.exit(5)
        return None

    def propfind(self, path):
        '''Do a PROPFIND request'''
        s = self.session
        s.auth = (self.user, self.token)
        try:
            r = s.request(
                method='PROPFIND',
                url=f'{self.http}://{self.host}/{path}',
                headers={'Depth': '0'},
                data=DIOS_MIO,
                verify=self.ssl
            )

            if r.status_code == 207:
                dom = minidom.parseString(r.text.encode('ascii', 'xmlcharrefreplace'))
                try:
                    return {
                        'last_modified': dom.getElementsByTagName(
                            'd:getlastmodified'
                        )[0].firstChild.data,
                        'content_type': dom.getElementsByTagName(
                            'd:getcontenttype'
                        )[0].firstChild.data,
                        'file_id': int(dom.getElementsByTagName(
                            'oc:fileid'
                        )[0].firstChild.data),
                        'size': int(dom.getElementsByTagName(
                            'oc:size'
                        )[0].firstChild.data),
                        'favorite': int(dom.getElementsByTagName(
                            'oc:favorite'
                        )[0].firstChild.data),
                        'owner': dom.getElementsByTagName(
                            'oc:owner-display-name'
                        )[0].firstChild.data,
                        'href': dom.getElementsByTagName(
                            'd:href'
                        )[0].firstChild.data
                    }
                except Exception:
                    # (markuman) I guess it's folder, because it has no content_type
                    return {
                        'last_modified': dom.getElementsByTagName(
                            'd:getlastmodified'
                        )[0].firstChild.data,
                        'content_type': 'inode/directory',
                        'file_id': dom.getElementsByTagName(
                            'oc:fileid'
                        )[0].firstChild.data,
                        'size': dom.getElementsByTagName(
                            'oc:size'
                        )[0].firstChild.data,
                        'favorite': dom.getElementsByTagName(
                            'oc:favorite'
                        )[0].firstChild.data,
                        'owner': dom.getElementsByTagName(
                            'oc:owner-display-name'
                        )[0].firstChild.data,
                        'href': dom.getElementsByTagName(
                            'd:href'
                        )[0].firstChild.data
                    }

            elif r.status_code == 404:
                return {}
            else:
                self.error(
                    {
                        "action": "propfind",
                        "message": f"Nextcloud instance returned status code: {r.status_code}",
                        "path": path,
                    }
                )
        except requests.exceptions.ReadTimeout as error:
            self.error(
                {
                    "action": "propfind",
                    "message": f"Timeout ({self.timeout} sec) error doing GET request.",
                    "error": error
                }
            )

    def put(self, path, src=None):
        '''Do a PUT request'''
        try:
            if src:
                r = self.session.put(
                    f'{self.http}://{self.host}/{path}',
                    data=open(src, 'rb'),
                    auth=(self.user, self.token),
                    verify=self.ssl,
                    timeout=self.timeout
                )
            else:
                r = self.session.put(
                    f'{self.http}://{self.host}/{path}',
                    headers=self.headers,
                    auth=(self.user, self.token),
                    verify=self.ssl,
                    timeout=self.timeout
                )

            if r.status_code in [200, 201, 204]:
                return r, True
            self.error(
                {
                    "action": "put",
                    "message": f"Nextcloud instance returned status code {r.status_code}",
                }
            )
        except requests.exceptions.ReadTimeout as error:
            self.error(
                {
                    "action": "put",
                    "message": f"Timeout ({self.timeout} sec) error doing GET request",
                    "error": error
                }
            )

    def delete(self, path):
        '''Do a DELETE request'''
        try:
            r = self.session.delete(
                f'{self.http}://{self.host}/{path}',
                auth=(self.user, self.token), verify=self.ssl, timeout=self.timeout
            )

            if r.status_code in [200, 204]:
                return r, True
            elif r.status_code == 404:
                return r, False
            else:
                self.error(
                    {
                        "action": "delete",
                        "message": f"Nextcloud instance returned status code {r.status_code}",
                        "path": path,
                    }
                )
        except requests.exceptions.ReadTimeout as error:
            self.error(
                {
                    "action": "delete",
                    "message": f"Timeout ({self.timeout} sec) error doing GET request",
                    "error": error
                }
            )

    def talk(self, message, channel):
        '''Post in Talk/Chat'''
        body = {
            'message': message,
            'replyTo': 0
        }

        spreed_v1_path = "ocs/v2.php/apps/spreed/api/v1/chat"
        try:
            r = self.session.post(
                f'{self.http}://{self.host}/{spreed_v1_path}/{channel}',
                data=body,
                headers=self.headers(),
                auth=(self.user, self.token),
                verify=self.ssl, timeout=self.timeout
            )

            if r.status_code == 201:
                return r, True
            else:
                self.error(
                    {
                        "action": "talk",
                        "message": f"Nextcloud instance returned status code {r.status_code}",
                        "talk_message": message,
                        "channel": channel
                    }
                )
        except requests.exceptions.ReadTimeout as error:
            self.error(
                {
                    "action": "talk",
                    "message": f"Timeout ({self.timeout} sec) error doing talk request",
                    "error": error
                }
            )
        return False

    def request_passwords_session(self):
        '''Request a Passwords API session'''
        return self.get("index.php/apps/passwords/api/1.0/session/request")

    def open_passwords_session(self, password_hash):
        '''Open a Passwords API session'''
        post_obj = {
            'challenge': password_hash
        }

        r = self.session.post(
            f'{self.http}://{self.host}/index.php/apps/passwords/api/1.0/session/open',
            data=post_obj,
            headers=self.headers(),
            auth=(self.user, self.token),
            verify=self.ssl, timeout=self.timeout
        )
        if r.status_code < 299:
            self.x_api_session = r.headers.get('X-API-SESSION')
            return r.json()
        else:
            self.error(
                {
                    "action": "open_password_session",
                    "message": f"Nextcloud instance returned status code {r.status_code}",
                    "password_hash": password_hash
                }
            )
        return False

    def close_passwords_session(self):
        '''Close Passwords API session'''
        return self.get("index.php/apps/passwords/api/1.0/session/close")

    def list_passwords(self):
        '''List all passwords'''
        if self.cache['last_update'] < (time.time() - self.cache_duration):
            self.debug(
                {
                    "action": "list_passwords",
                    "message": "Updating cached list of password",
                    "last_update": self.cache['last_update'],
                    "cache_duration": self.cache_duration
                }
            )
            self.cache['cached_passwords'] = self.get(
                "index.php/apps/passwords/api/1.0/password/list"
            )
            if self.cache['cached_passwords']:
                self.cache['last_update'] = time.time()
                self._write_cache()
        else:
            self.debug(
                { "action": "list_passwords", "message": "Reusing cached list of password" }
            )
        return self.cache['cached_passwords']

    def list_passwords_folders(self):
        '''List passwords folders'''
        return self.get("index.php/apps/passwords/api/1.0/folder/list")

    def delete_passwords_folder(self, name):
        '''Delete a passwords folder'''
        all_folders = self.list_passwords_folders()
        folder_id = None
        for folder in all_folders:
            if folder['label'] == name:
                folder_id = folder['id']
                self.debug({
                    "action": "delete_passwords_folder",
                    "message": "found folder to delete",
                    "object": folder['id']
                })
        if folder_id:
            post_obj = {
                'id': folder_id
            }
            r = self.session.delete(
                f'{self.http}://{self.host}/index.php/apps/passwords/api/1.0/folder/delete',
                data=post_obj,
                headers=self.headers(),
                auth=(self.user, self.token),
                verify=self.ssl,
                timeout=self.timeout
            )
            if r.status_code == 201:
                return r.json()
            self.error(
                {
                    "action": "delete_passwords_folder",
                    "message": f"Nextcloud instance returned status code {r.status_code}",
                    "folder_name": name
                }
            )
        else:
            self.error(
                {
                    "action": "delete_passwords_folder",
                    "message": "Folder not found",
                    "folder_name": name
                }
            )
        return False

    def exists_passwords_folder(self, name):
        '''Test if a passwords folder exists'''
        for folder in self.list_passwords_folders():
            if folder.get('label') == name:
                return folder
        return False

    def create_passwords_folder(self, name):
        '''Create passwords folder'''
        if not self.exists_passwords_folder(name):
            post_obj = {
                'label': name
            }

            r = self.session.post(
                f'{self.http}://{self.host}/index.php/apps/passwords/api/1.0/folder/create',
                data=post_obj,
                headers=self.headers(),
                auth=(self.user, self.token),
                verify=self.ssl,
                timeout=self.timeout
            )

            if r.status_code == 201:
                return r.json()
            self.error(
                {
                    "action": "create_passwords_folder",
                    "message": f"Nextcloud instance returned status code {r.status_code}",
                    "folder_name": name
                }
            )
        else:
            self.warning(
                {
                    "action": "create_passwords_folder",
                    "message": "Passwords folder already exists",
                    "folder_name": name
                }
            )
        return False

    def get_passwords_folder(self, name):
        '''Get Passwords folder'''
        for folder in self.list_passwords_folders():
            if folder.get('label') == name:
                return folder.get('id')
        return None

    def get_password(self, name, details):
        '''Get a password'''
        r = self.list_passwords()
        ret = []
        for item in r:
            if item['cseType'] == 'CSEv1r1' and self.e2e_password:
                item_key = self.keychain['keys'].get(item['cseKey'])
                if item_key:
                    if decrypt_item(item_key, item, 'label') == name:
                        if details:
                            item['password'] = decrypt_item(item_key, item, 'password')
                            item['url'] = decrypt_item(item_key, item, 'url')
                            item['username'] = decrypt_item(item_key, item, 'username')
                            item['label'] = decrypt_item(item_key, item, 'label')
                            item['notes'] = decrypt_item(item_key, item, 'notes')
                            item['customFields'] = decrypt_item(item_key, item, 'customFields')
                            ret.append(item)
                        else:
                            ret.append(decrypt_item(item_key, item, 'password'))
            else:
                if item['label'] == name:
                    if details:
                        ret.append(item)
                    else:
                        ret.append(item['password'])
        return ret

    def fetch_generated_password(self):
        '''Fetch a generated password'''
        return self.get('index.php/apps/passwords/api/1.0/service/password').get('password')

    def exists_password(self, obj):
        '''Test if a password exist with the same name'''
        safer_obj = dict(obj, **{ 'password': '***' })
        self.debug(
            { "action": "check_exists_password", "object": safer_obj }
        )
        all_passwords = self.list_passwords()
        if all_passwords:
            for password in all_passwords:
                if self.is_same_password(obj, password):
                    safer_password = dict(password, **{ 'password': '***' })
                    self.debug(
                        {
                            "action": "test_exists_password",
                            "message": "Same password already exists",
                            "object": safer_obj,
                            "existing_password": safer_password
                        }
                    )
                    return password
        self.debug(
            {
                "action": "test_exists_password",
                "message": "No password match found",
                "object": safer_obj
            }
        )
        return False

    def create_password(self, post_obj, update):
        '''Create/add a password'''
        new_obj = {}
        for key in post_obj.keys():
            if key in self.field_replacements:
                new_obj[self.field_replacements[key]] = post_obj[key]
            else:
                new_obj[key] = post_obj[key]
        self.debug(
            {
                "action": "create_password",
                "message": "Result of replacing fields",
                "original_object": post_obj,
                "new_object": new_obj
            }
        )
        if new_obj.get('folder', '') != '':
            folder_id = self.get_folder_id(new_obj['folder'])
            if not folder_id:
                folder_id = self.create_passwords_folder(new_obj['folder'])['id']
            new_obj['folder'] = folder_id
        else:
            new_obj.pop('folder', None)
        if not 'username' in new_obj:
            new_obj['username'] = ''
        if not 'url' in new_obj:
            new_obj['url'] = ''
        exists_password = self.exists_password(new_obj)
        if not exists_password:
            try:
                self.debug(
                    {
                        "action": "create_password",
                        "object": new_obj
                    }
                )
                r = self.session.post(
                    f'{self.http}://{self.host}/index.php/apps/passwords/api/1.0/password/create',
                    data=new_obj,
                    headers=self.headers(),
                    auth=(self.user, self.token),
                    verify=self.ssl, timeout=self.timeout
                )

                if r.status_code == 201:
                    if self.cache['cached_passwords']:
                        self.cache['cached_passwords'].append(new_obj)
                    else:
                        self.cache['cached_passwords'] = [ new_obj ]
                    return r.json()
                self.error(r.json())
                self.error(
                    {
                        "action": "create_password",
                        "message": f"Nextcloud instance returned status code: {r.status_code}",
                        "object": new_obj
                    }
                )
            except requests.exceptions.ReadTimeout as error:
                self.error(
                    {
                        "action": "create_password",
                        "message": f"Timeout ({self.timeout} sec) error doing GET request",
                        "object": new_obj,
                        "error": error
                    }
                )
        else:
            if update:
                new_obj['id'] = exists_password['id']
                return self.update_password(new_obj)
            self.warning(
                {
                    "action": "create_password",
                    "message": "Password with that name/label already exists",
                    "object": new_obj,
                    "label": new_obj['label']
                }
            )
        return False

    def delete_password(self, post_obj):
        '''Delete a password'''
        try:
            r = self.session.delete(
                f'{self.http}://{self.host}/index.php/apps/passwords/api/1.0/password/delete',
                data=post_obj,
                headers=self.headers(),
                auth=(self.user, self.token),
                verify=self.ssl, timeout=self.timeout
            )

            if r.status_code < 299:
                return r.json()
            self.error(
                {
                    "action": "delete_password",
                    "message": f"Nextcloud instance returned status code {r.status_code}",
                }
            )
        except requests.exceptions.ReadTimeout as error:
            self.error(
                {
                    "action": "delete_password",
                    "message": f"Timeout ({self.timeout} sec) error doing GET request",
                    "error": error,
                }
            )
        return False


    def update_password(self, post_obj):
        '''Update a password'''
        safer_obj = dict(post_obj, **{ 'password': '***' })
        try:
            r = self.session.patch(
                f'{self.http}://{self.host}/index.php/apps/passwords/api/1.0/password/update',
                data=post_obj,
                headers=self.headers(),
                auth=(self.user, self.token),
                verify=self.ssl, timeout=self.timeout
            )

            if r.status_code < 299:
                self.debug(
                    {
                        "action": "update_password",
                        "message": "Updated password",
                        "object": safer_obj,
                        "status_code": r.status_code
                    }
                )
                return r.json()
            self.error(
                {
                    "action": "update_password",
                    "message": f"Nextcloud instance returned status code {r.status_code}",
                    #"content": r.content,
                    "json": r.json(),
                    "headers": r.headers,
                    # "request_headers": r.request.headers,
                    "object": safer_obj,
                }
            )
        except requests.exceptions.ReadTimeout as error:
            self.error(
                {
                    "action": "update_password",
                    "message": f"Timeout ({self.timeout} sec) error doing GET request",
                    "object": safer_obj,
                    "error": error,
                }
            )
        return False

    def is_same_key(self, key, dict1, dict2):
        '''Test if two dictionaries have the same key with the same value'''
        if (
            (key in dict1 or key.lower() in dict1) and
            (key in dict2 or key.lower() in dict2)
            ):
            if (
                dict1[key] == dict2[key] or
                dict1[key.lower()] == dict2[key] or
                dict1[key.lower()] == dict2[key.lower()] or
                dict1[key] == dict2[key.lower()]
                ):
                return True
        return False

    def is_same_password(self, obj1, obj2):
        '''Test if two password objects are the same or similar'''
        safer_obj1 = dict(obj1, **{ 'password': '***' })
        safer_obj2 = dict(obj2, **{ 'password': '***' })
        if obj1 == obj2:
            self.debug(
                {
                    "action": "notify_exact_match",
                    "object": {
                        "obj1": safer_obj1,
                        "obj2": safer_obj2
                    }
                }
            )
            return True
        # if obj1['label'] == obj2['label']:
            # self.debug(
            #     {
            #         "action": "notify_name_match",
            #         "object": {
            #             "obj1": safer_obj1,
            #             "obj2": safer_obj2
            #         }
            #     }
            # )
        if self.is_same_key('username', obj1, obj2):
            if self.is_same_key('password', obj1, obj2):
                if self.is_same_key('url', obj1, obj2):
                    #if self.is_same_key('folder', obj1, obj2):
                    self.debug(
                        {
                            "action": "notify_match",
                            "object": {
                                "obj1": safer_obj1,
                                "obj2": safer_obj2
                            }
                        }
                    )
                    return True
        return False

    def get_folder_id(self, folder_name):
        '''Get a folder id from the name'''
        for folder in self.list_passwords_folders():
            if folder['label'] == folder_name:
                return folder['id']
        return False

class NcPasswordClient:
    '''Nextcloud Password Client'''

    def __init__(
        self, debug_level, log_file, host, user, api_token, cse_password,
        timeout, cache_duration, https_proxy, output_format, field_replacements
        ):
        self.config = {}
        self.output_format = output_format
        self.config['debug_level'] = debug_level
        if log_file is None:
            self.config['log_file'] = os.path.join(
                os.environ.get(
                    'HOME',
                    os.environ.get(
                        'USERPROFILE',
                        os.getcwd()
                    )
                ),
                'log',
                'nc_password_client.log'
            )
        self._init_log()
        params = {
            "host": host,
            "user": user,
            "api_token": api_token,
            "cse_password": cse_password,
            "timeout": timeout,
            "cache_duration": cache_duration,
            "https_proxy": https_proxy,
            "logger": self._log,
            "output_format": output_format,
            "field_replacements": field_replacements,
        }
        self.nc = NextcloudHandler(params)

    def _safer_obj(self, obj, fields=None):
        if fields is None:
            fields = ['password', 'token', 'secret', 'pass', 'passwd', 'hash']
        if isinstance(obj, dict):
            for key in obj.keys():
                if isinstance(obj[key], dict) or isinstance(obj[key], list):
                    obj[key] = self._safer_obj(obj[key], fields=fields)
                else:
                    if key in fields:
                        obj[key] = '***'
        if isinstance(obj, list):
            for item in obj:
                item = self._safer_obj(item, fields=fields)
        return obj

    def _output(self, obj, unsafe=False):
        if unsafe:
            out_obj = obj
        else:
            out_obj = self._safer_obj(obj)
        if self.output_format == 'json':
            return json.dumps(out_obj, indent=2)
        if self.output_format == 'yaml':
            return dump(out_obj, Dumper=Dumper)
        output = ''
        if isinstance(out_obj, list):
            output += '-' * os.get_terminal_size(0)[0]
            for item in obj:
                output += f"{self._output(item)}\n"
        elif isinstance(out_obj, dict):
            output += '-' * os.get_terminal_size(0)[0]
            for key in out_obj.keys():
                output += f"{key} = {self._output(out_obj[key])}\n"
        else:
            output += f"{out_obj}\n"
        return output.strip('\n')

    def debug(self, obj, unsafe=False):
        '''Show debug information'''
        self._log.debug(
            self._output(obj, unsafe=unsafe)
        )

    def warning(self, obj, unsafe=False):
        '''Show warning information'''
        self._log.warning(
            self._output(obj, unsafe=unsafe)
        )

    def info(self, obj, unsafe=False):
        '''Show information'''
        self._log.info(
            self._output(obj, unsafe=unsafe)
        )

    def error(self, obj, unsafe=False):
        '''Show error information'''
        try:
            self._log.error(
                self._output(obj, unsafe=unsafe)
            )
        except TypeError as error:
            self._log.error(
                'Additional error showing the error from %s. %s',
                obj,
                error
            )

    def get_password(self, name, details):
        '''Get a single password'''
        self.info(self.nc.get_password(name, details))

    def list_passwords(self):
        '''List all passwords'''    
        self.info(self.nc.list_passwords())

    def list_passwords_folders(self):
        '''List all password folders'''    
        self.info(self.nc.list_passwords_folders())

    def create_password(
        self, obj, update, label=None, password=None, folder=None, username=None,
        url=None, notes=None, custom_fields=None, sha1_hash=None, hidden=None, favorite=None):
        '''Create a password with an object'''
        if obj is None:
            obj = {}
        if label:
            obj['label'] = label
        if password:
            obj['password'] = password
        if username:
            obj['username'] = username
        if url:
            obj['url'] = url
        if notes:
            obj['notes'] = notes
        if custom_fields:
            obj['custom_fields'] = custom_fields
        if hidden is not None:
            obj['hidden'] = hidden
        if favorite is not None:
            obj['favorite'] = favorite
        if folder:
            obj['folder'] = folder
        if sha1_hash:
            obj['sha1_hash'] = sha1_hash
        safer_obj = dict(obj, **{ 'password': '***' })
        self.debug(
            { "action": "create_password", "object": safer_obj }
        )
        self.debug(
            {
                "action": "created_password",
                "object": self.nc.create_password(obj, update)
            }
        )

    def delete_password(self, name, password_id):
        '''Delete a password'''
        if name is not None:
            field = 'label'
        elif password_id is not None:
            field = 'id'
        else:
            self.error(
                {
                    "action": "delete_password",
                    "message": "You must indicate a name or a password id to delete"
                }
            )
            sys.exit(2)
        for password in self.nc.list_passwords():
            if password[field] == name:
                safer_obj = dict(password, **{ 'password': '***' })
                self.debug(
                    { "action": "delete_password", "object": safer_obj }
                )
                self.debug(
                    { "action": "deleted_password", "object": self.nc.delete_password(password)}
                )
                return True
        self.warning(
            {
                "action": "delete_password",
                "message": f"Password with that {field} doesn't exist",
                "object": name
            }
        )
        return False

    def create_passwords_folder(self, name):
        '''Create a passwords folder'''
        self.info(self.nc.create_passwords_folder(name))

    def delete_passwords_folder(self, name):
        '''Delete a passwords folder'''
        self.info(self.nc.delete_passwords_folder(name))

    def migrate_pass(self, limit=-1, update=False):
        '''Migrate password store to Nextcloud pass'''
        store = passpy.store.Store()
        count = 0
        for item in store.find(''):
            if limit == -1 or limit > count:
                obj = {
                    "label": os.path.basename(item),
                }
                folder = os.path.dirname(item)
                if folder != '':
                    obj['folder'] = folder
                raw_values = store.get_key(item).split('\n')
                obj['password'] = raw_values[0]
                raw_values.pop(0)
                custom_fields = {}
                for line in raw_values:
                    if ': ' in line:
                        split_line = line.split(': ')
                        field = split_line[0].lower()
                        value = split_line[1]
                    else:
                        field = line
                        value = ''
                    if value != '':
                        if field == '':
                            field = 'notes'
                        if field == 'login':
                            field = 'username'
                        if field == 'uri':
                            field = 'url'
                        if field not in custom_fields:
                            custom_fields[field] = value
                        else:
                            if isinstance(custom_fields[field], str):
                                custom_fields[field] = [ custom_fields[field], value]
                            else:
                                custom_fields[field].append(value)
                        obj[field] = value
                obj['customFields'] = json.dumps(custom_fields)
                new_password = self.nc.create_password(obj, update=update)
                if new_password:
                    self.debug(
                        { "action": "created_password", "object": new_password }
                    )
                    count += 1
        self.debug(
            {
                "action": "migrate_pass",
                "message": "Migrate Pass summary",
                "limi": limit,
                "update": update,
                "migrated_passwords": count
            }
        )

    def delete_all_passwords(self, yes_i_am_sure):
        '''DANGEROUS! Delete ALL passwords from your Nextcloud Password instance'''
        if not yes_i_am_sure:
            answer = input(
                'WARNING! Are you completely sure you want to delete ALL your passwords? (Y/n)'
            )
            if answer != "Y":
                self.info(
                    {
                        "action": "delete_all_passwords",
                        "message": "Aborting after answering something different from 'Y'"
                    }
                )
                return False
        for item in self.nc.list_passwords():
            safer_obj = dict(item, **{ 'password': '***' })
            self.debug(
                {
                    "action": "delete_all_passwords",
                    "message": "Deleting password",
                    "object": safer_obj
                }
            )
            self.nc.delete_password(item)
        return True

    def remove_duplicates(self, limit):
        '''Remove duplicate passwords'''
        checked_passwords = []
        count = 0
        if limit == 0:
            return True
        passwords = self.nc.list_passwords()
        if passwords is None:
            return False
        for item in passwords:
            for checked in checked_passwords:
                if self.nc.is_same_password(checked, item):
                    self.debug(
                        {
                            "action": "remove_duplicate",
                            "object": self.nc.delete_password(item)
                        }
                    )
                    count += 1
                    break
            checked_passwords.append(item)
            if limit != -1 and count >= limit:
                break
        self.debug(
            {
                "action": "remove_duplicates",
                "message": "Remove duplicates summary",
                "checked_password": len(checked_passwords),
                "removed_duplicates": count
            }
        )
        return True

    def search(self, pattern, limit):
        '''Search for a password (or passwords) matching in any field the given pattern'''
        passwords = self.nc.list_passwords()
        if passwords is None:
            self.info(
                {
                    "action": "search",
                    "message": "No passwords obtained"
                }
            )
            return False
        count = 0
        for password in passwords:
            for key in password.keys():
                match = re.search(
                    pattern,
                    f"{password[key]}",
                    re.IGNORECASE | re.MULTILINE
                )
                if match:
                    self.info(
                        password,
                        unsafe=True
                    )
                    count += 1
                    if limit != -1 and count >= limit:
                        return True
                    break
        if count == 0:
            self.info(
                {
                    "action": "search",
                    "message": "No passwords found with the given pattern",
                    "patter": pattern,
                    "total_passwords": len(passwords)
                }
            )
        return False

    def import_csv(
        self, limit, update, csv_file, encoding='utf-8',
        delimiter=';', quotechar='"', first_line_fields=True, fieldnames=None
        ):
        '''Import passwords from a CSV file'''
        self.debug(
            {
                "action": "import_csv",
                "message": "Importing from CSV file",
                "csv_file": csv_file
            }
        )
        if first_line_fields:
            with open(csv_file, encoding=encoding) as file_object:
                reader = csv.reader(file_object, delimiter=delimiter, quotechar=quotechar)
                for field_names in reader:
                    self.debug(
                        {
                            "action": "import_csv",
                            "message": "Fields in the first row",
                            "field_names": field_names
                        }
                    )
                    break
        else:
            if fieldnames is None:
                self.error(
                    {
                        "action": "import_csv",
                        "message": "You must indicate a list of fieldnames or first-line-fields"
                    }
                )
                sys.exit(4)
            field_names = fieldnames.split(',')
        with open(csv_file, encoding=encoding) as file_object:
            reader = csv.DictReader(
                file_object, fieldnames=field_names,
                delimiter=delimiter, quotechar=quotechar
            )
            count = 0
            r_count = 0
            for row in reader:
                if r_count > 0:
                    password = row
                    self.debug(
                        {
                            "action": "import_csv",
                            "message": "Importing password",
                            "object": password
                        }
                    )
                    if self.create_password(password, update=update):
                        count += 1
                        if count == limit:
                            break
                r_count += 1
        self.debug(
            {
                "action": "import_csv",
                "message": "Importation summary",
                "rows_in_file": r_count,
                "imported_password": count
            }
        )
        return True

    def _init_log(self):
        ''' Initialize log object '''
        self._log = logging.getLogger("nc_password_client")
        self._log.setLevel(logging.DEBUG)

        streamhandler = logging.StreamHandler(sys.stdout)
        streamhandler.setLevel(
            logging.getLevelName(self.config.get("debug_level", 'INFO'))
        )
        self._log.addHandler(streamhandler)
        return True


@click.group()
@click.option(
    "--debug-level",
    "-d",
    default="INFO",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        case_sensitive=False,
    ),
    help='Set the debug level for the standard output.'
)
@click.option(
    "--output-format",
    "-f",
    default="json",
    type=click.Choice(
        ["json", "yaml", "plain"],
        case_sensitive=False,
    ),
    help='Set the output format.'
)
@click.option('--log-file', '-l', help="File to store all debug messages.")
# @click.option("--dummy","-n", is_flag=True,
# help="Don't do anything, just show what would be done.")
@click.option(
    '--host', '-H',
    required=True,
    help='Nextcloud host name'
)
@click.option(
    '--user', '-u',
    required=True,
    help='Nextcloud user name'
)
@click.option(
    '--api-token', '-a',
    required=True,
    help='Nextcloud user\'s application token'
)
@click.option(
    '--cse-password', '-p',
    help='Nextcloud user\'s end-to-end encryption password'
)
@click.option(
    '--timeout', '-t',
    default=10,
    type=click.INT,
    help='Timeout for HTTPS requests'
)
@click.option(
    '--cache-duration', '-c',
    default=300,
    type=click.INT,
    help='Number of seconds to hold the list of passwords'
)
@click.option(
    '--https-proxy', '-P',
    help='HTTPS proxy to use to connect to the Nextcloud instance'
)
@click.option(
    '--field-replacements', '-F',
    multiple=True,
    default=[
        "login:username",
        "uri:url",
        "pass:password",
        "name:label",
        "user_name:username",
        "web:url",
    ],
    help='Pairs of field names to be replaced separated by a : (colon). Like: -F login:username'
)
@click_config_file.configuration_option()
@click.pass_context
def cli(
    ctx, debug_level, log_file, host, user, api_token, cse_password,
    timeout, cache_duration, https_proxy, output_format, field_replacements
    ):
    '''Client function to pass context'''
    ctx.ensure_object(dict)
    ctx.obj['NcPasswordClient'] = NcPasswordClient(
        debug_level, log_file, host, user, api_token, cse_password,
        timeout, cache_duration, https_proxy, output_format, field_replacements
    )

@cli.command()
@click.option('--name', '-n', required=True, help='Name of the password to show')
@click.option(
    '--details', '-D',
    is_flag=True,
    default=False,
    help='Show all details of the password or just the password (False)'
)
@click_config_file.configuration_option()
@click.pass_context
def show(ctx, name, details):
    '''Show a single password'''
    ctx.obj['NcPasswordClient'].get_password(name, details)

@cli.command()
@click_config_file.configuration_option()
@click.pass_context
def ls(ctx):
    '''List all password'''
    ctx.obj['NcPasswordClient'].list_passwords()

@cli.command()
@click.option(
    '--obj', '-o',
    required=False,
    default='',
    help='JSON object for a password'
)
@click.option(
    '--label', '-l',
    required=False,
    default='',
    help='Label/name of the password'
)
@click.option(
    '--username', '-u',
    required=False,
    default='',
    help='User name of the password'
)
@click.option(
    '--password', '-p',
    required=False,
    default='',
    help='Password or secret to store'
)
@click.option(
    '--url', '-U',
    required=False,
    default='',
    help='The url of the associated website'
)
@click.option(
    '--notes', '-n',
    required=False,
    default='',
    help='The users notes'
)
@click.option(
    '--custom-fields', '-c',
    required=False,
    default='',
    help='The custom fields defined by the user'
)
@click.option(
    '--sha1-hash', '-H',
    required=False,
    default='',
    help='The SHA1 hash of the password'
)
@click.option(
    '--folder', '-f',
    required=False,
    default='',
    help='The current folder of the password'
)
@click.option(
    '--hidden', '-i',
    required=False,
    is_flag=True,
    default=False,
    help='Whether or not the password should be hidden'
)
@click.option(
    '--favorite', '-f',
    required=False,
    is_flag=True,
    default=False,
    help='Whether or not the user has marked this password as favorite'
)
@click.option(
    '--update', '-u',
    is_flag=True,
    default=False,
    help='Update existing passwords with the data from Password Store'
)
@click_config_file.configuration_option()
@click.pass_context
def create_password(ctx, obj, update, label, password, folder, username,
    url, notes, custom_fields, sha1_hash, hidden, favorite):
    '''Create a password'''
    ctx.obj['NcPasswordClient'].create_password(
        json.loads(obj), update, label, password, folder, username,
        url, notes, custom_fields, sha1_hash, hidden, favorite
    )

@cli.command()
@click.option('--name', '-n', help='Name of the password to delete')
@click.option('--password-id', '-i', help='ID of the password to delete')
@click_config_file.configuration_option()
@click.pass_context
def delete_password(ctx, name, password_id):
    '''Delete a password'''
    ctx.obj['NcPasswordClient'].delete_password(name, password_id)

@cli.command()
@click.option('--name', '-n', required=True, help='Name of the passwords folder to create')
@click_config_file.configuration_option()
@click.pass_context
def create_passwords_folder(ctx, name):
    '''Create a password folder'''
    ctx.obj['NcPasswordClient'].create_passwords_folder(name)

@cli.command()
@click.option(
    '--limit', '-l',
    default=-1,
    help='Maximun number of passwords to remove. -1 for unlimited.'
)
@click_config_file.configuration_option()
@click.pass_context
def remove_duplicates(ctx, limit):
    '''Remove duplicate passwords'''
    ctx.obj['NcPasswordClient'].remove_duplicates(limit)

@cli.command()
@click_config_file.configuration_option()
@click.pass_context
def list_passwords_folders(ctx):
    '''List all password folders'''
    ctx.obj['NcPasswordClient'].list_passwords_folders()

@cli.command()
@click.option('--name', '-n', required=True, help='Name of the passwords folder to delete')
@click_config_file.configuration_option()
@click.pass_context
def delete_passwords_folder(ctx, name):
    '''Delete a password folder'''
    ctx.obj['NcPasswordClient'].delete_passwords_folder(name)

@cli.command()
@click.option(
    '--pattern', '-p',
    required=True,
    help='Regular expression pattern to search in the password fields'
)
@click.option(
    '--limit', '-l',
    default=-1,
    help='Maximun number of passwords to show. -1 for unlimited.'
)
@click_config_file.configuration_option()
@click.pass_context
def search(ctx, pattern, limit):
    '''Search for a password (or passwords) that match the given pattern in its fields'''
    ctx.obj['NcPasswordClient'].search(pattern, limit)

@cli.command()
@click.option(
    '--limit', '-l',
    default=-1,
    help='Maximun number of passwords to migrate. -1 for unlimited.'
)
@click.option(
    '--update', '-u',
    is_flag=True,
    default=False,
    help='Update existing passwords with the data from Password Store'
)
@click_config_file.configuration_option()
@click.pass_context
def migrate_pass(ctx, limit, update):
    '''Migrate Password Store passwords to Nextcloud Passwords'''
    ctx.obj['NcPasswordClient'].migrate_pass(limit, update)

@cli.command()
@click.option(
    '--yes-I-am-sure',
    default=False,
    help="DAGEROUS! Don't prompt for confirmation before deleting ALL passwords."
)
@click_config_file.configuration_option()
@click.pass_context
def delete_all_passwords(ctx, yes_i_am_sure):
    '''Delete all passwords'''
    ctx.obj['NcPasswordClient'].delete_all_passwords(yes_i_am_sure)

@cli.command()
@click.option(
    '--limit', '-l',
    default=-1,
    help='Maximun number of passwords to migrate. -1 for unlimited.'
)
@click.option(
    '--update', '-u',
    is_flag=True,
    default=False,
    help='Update existing passwords with the data from Password Store'
)
@click.option(
    '--csv-file', '-c',
    required=True,
    type=click.Path(exists=True),
    help='CSV file with your passwords to import'
)
@click.option(
    '--encoding', '-e',
    default='utf-8',
    help='Character encoding of the CSV file'
)
@click.option(
    '--delimiter', '-d',
    default=';',
    help='Character delimiter of columns in the rows of the CSV file'
)
@click.option(
    '--quotechar', '-q',
    default='"',
    help='Character to quote strings in the rows of the CSV file'
)
@click.option(
    '--first-line-fields', '-f',
    is_flag=True,
    default=True,
    help='The first line contains the field names'
)
@click.option(
    '--fieldnames', '-F',
    default='name,username,password,url',
    help='Field names separated by commans in case the first line does NOT contain the field names'
)
@click_config_file.configuration_option()
@click.pass_context
def import_csv(
    ctx, limit, update, csv_file, encoding,
    delimiter, quotechar, first_line_fields, fieldnames):
    '''Import passwords stored in a CSV file to Nextcloud Passwords'''
    ctx.obj['NcPasswordClient'].import_csv(
        limit, update, csv_file, encoding,
        delimiter, quotechar, first_line_fields, fieldnames)

if __name__ == "__main__":
    cli(obj={})
